/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.w3c.rdf.util;

import org.w3c.rdf.model.*;
import org.w3c.rdf.model.impl.*;
import org.w3c.rdf.util.SetOperations;
import org.w3c.tools.crypt.*;
import org.w3c.rdf.syntax.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.*;
import org.w3c.rdf.vocabulary.rdf_schema_19990303.*;
import org.xml.sax.*;

import java.util.*;
import java.io.*;
import java.net.*;

import java.security.*;

/**
 * Useful utility methods.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class RDFUtil {

  // implemented in Resource constructor
  //  static long nonameCount = 0;
  static Random rnd;
  
  /**
   * Creates a new unique unnamed resource.
   */
  public static Resource noname() {

    if(rnd == null) {
      rnd = new SecureRandom();
    }
    byte[] b = new byte[16];
    rnd.nextBytes(b);
    return new ResourceImpl("urn:rdf:noname:" + DigestUtil.toHexString(b));
  }


  /*
  public static Resource noname(String namespace) {
    return new Resource(namespace + "#" + Integer.toHexString( rnd.nextInt() ));
  }
  */

  static final int MAX_ORD = 30; // maximal number of cached ordinals
  static Resource[] ords;
  static {
    ords = new Resource[MAX_ORD + 1]; // 0 is wasted
    for(int i=1; i <= MAX_ORD; i++)
      ords[i] = new ResourceImpl(RDF._Namespace + "_" + i);
  }

  /**
   * Creates an ordinal property (rdf:li, rdf:_N)
   */
  public static Resource ord(int i) {
    
    if(i < 1)
      throw new IllegalArgumentException("Attempt to construct invalid ordinal resource");
    if(i <= MAX_ORD)
      return ords[i];
    else
      return new ResourceImpl(RDF._Namespace + "_" + i);
  }


  /**
   * Converts an ordinal property to an integer.
   */
  public static int getOrd(Resource r) {

    if(r == null)
      return -1;

    String uri = r.toString();
    int pos = getNamespaceEnd(uri);
    if(pos > 0 && pos + 1 < uri.length() && isRDF(uri)) {
      try {
	int n = Integer.parseInt(uri.substring(pos + 1));
	if(n >= 1)
	  return n;
      } catch (Exception any) {}
    }
    return -1;
  }

  /**
   * Tests if the URI is qualified, i.e. has a namespace prefix.
   */
  public static boolean isQualified(String s) {
    
    int l = s.length()-1;
    do {
      char c = s.charAt(l);
      if(c == '#' || c == ':')
	return true;
      l--;
    } while (l >= 0);
    return false;
  }

  /**
   * Extracts the namespace prefix out of a URI.
   */
  public static String guessNamespace(String uri) {

    int l = getNamespaceEnd(uri);
    return l > 1 ? uri.substring(0, l /*-1*/) : "";
  }

  /**
   * Delivers the name out of the URI (without the namespace prefix).
   */
  public static String guessName(String uri) {

    return uri.substring(getNamespaceEnd(uri));
  }

  /**
   * Tests if the URI belongs to the RDF syntax/model namespace.
   */
  public static boolean isRDF(String uri) {
    return uri != null && uri.startsWith(RDF._Namespace);
  }

  /**
   * Tests if the resource belongs to the RDF syntax/model namespace.
   */
  public static boolean isRDF(Resource r) {
    return isRDF(r.getURI());
  }

  /**
   * Position of the namespace end
   */
  static int getNamespaceEnd(String uri) {

    int l = uri.length()-1;
    do {
      char c = uri.charAt(l);
      if(c == '#' || c == ':' || c == '/')
	break;
      l--;
    } while (l >= 0);
    l++;
    return l;
  }

  /**
   * Returns the first triple of the model
   */
  public static Statement get1(Model m) {
    if(m == null || m.size() == 0)
      return null;
    //    if(m.size() > 1)
    //      throw new RuntimeException("Model contains more than one triple");
    return (Statement)m.elements().nextElement();
  }

  /**
   * Prints the triples of a model to the given PrintStream.
   */
  public static void printStatements(Model m, PrintStream ps) {

    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement t = (Statement)en.nextElement();
      ps.println (t); // "triple(\""+t.subject()+"\",\""+t.predicate()+"\",\""+t.object()+"\").");
    }
  }

  /**
   * Dumps the model in a serialized form.
   */
  public static void dumpModel(Model m, PrintStream ps, RDFSerializer s) throws Exception {
    s.serialize( m, new OutputStreamWriter(ps) );
  }

  /**
   * Dumps the model in a serialized form in a string
   */
  public static String dumpModel(Model m, RDFSerializer s) throws Exception {

    StringWriter w = new StringWriter();
    s.serialize( m, w );
    return w.toString();
  }

  /**
   * Collects the triples of a model in a vector.
   */
  public static Vector getStatementVector(Model m) {

    Vector v = new Vector();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement t = (Statement)en.nextElement();
      v.addElement(t);
    }
    return v;
  }

  /**
   * Removes all triples which have something to do with the given namespace
   */
  public static Model removeNamespace( String ns, Model m ) {

    Model res = m.duplicate();
    for(Enumeration en = m.duplicate().elements(); en.hasMoreElements();) {

      Statement t = (Statement)en.nextElement();
      if(t.subject().toString().startsWith( ns ) ||
	 t.predicate().toString().startsWith( ns ) ||
	 t.object().toString().startsWith( ns )) {
	//	System.err.println("REMOVING TRIPLE: " + t);
	res.remove(t);
      }
    }
    return res;
  }

  /**
   * returns a subgraph of "m" containing
   * "r" and all nodes reachable from "r" via directed edges.
   * These edges are also included in the resulting model.
   */
  public static Model getReachable(Resource r, Model m) {

    Model result = m.create();
    getReachable(r, m, result);
    return result;
  }

  // FIXME: use digest instead of size
  static void getReachable(Resource r, Model m, Model result) {

    int oldSize = result.size();
    Model directlyReachable = m.find(r, null, null);
    SetOperations.unite(result, directlyReachable);
    if(result.size() == oldSize)
      return;
    for(Enumeration en = directlyReachable.elements(); en.hasMoreElements();) {
      Statement t = (Statement)en.nextElement();
      if(t.object() instanceof Resource)
	getReachable((Resource)t.object(), m, result);
    }
  }

  public static void parse(String fileNameOrURL, RDFParser parser, Model model) throws IOException, SAXException, MalformedURLException {

    URL url = new URL (normalizeURI(fileNameOrURL));

    // maybe this model is loaded as schema...
    //    Model model = factory.registry().get(url.toString());
    //    if(model != null)
    //      return model;

    // Prepare input source
    InputSource source = new InputSource( url.openStream() );
    source.setSystemId( url.toString() );
    model.setSourceURI( url.toString() );

    parser.parse( source, new ModelConsumer(model) );
  }

  public static String normalizeURI(String uri) {

    // normalise uri
    URL url = null;
    try {
      url = new URL (uri);
    } catch (Exception e) {
      try {
	if(uri.indexOf(':') == -1)
	  url = new URL ("file", null, uri);
      } catch (Exception e2) {}
    }
    return url != null ? url.toString() : uri;
  }


}
